# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from ceilometerclient.apiclient import client
from ceilometerclient.apiclient import fake_client
from ceilometerclient.tests.unit import utils
import ceilometerclient.v2.samples

GET_OLD_SAMPLE = {u'counter_name': u'instance',
                  u'user_id': u'user-id',
                  u'resource_id': u'resource-id',
                  u'timestamp': u'2012-07-02T10:40:00',
                  u'source': u'test_source',
                  u'message_id': u'54558a1c-6ef3-11e2-9875-5453ed1bbb5f',
                  u'counter_unit': u'',
                  u'counter_volume': 1.0,
                  u'project_id': u'project1',
                  u'resource_metadata': {u'tag': u'self.counter',
                                         u'display_name': u'test-server'},
                  u'counter_type': u'cumulative'}
CREATE_SAMPLE = copy.deepcopy(GET_OLD_SAMPLE)
del CREATE_SAMPLE['message_id']
del CREATE_SAMPLE['source']
CREATE_LIST_SAMPLE = copy.deepcopy(CREATE_SAMPLE)
CREATE_LIST_SAMPLE['counter_name'] = 'image'

GET_SAMPLE = {
    "user_id": None,
    "resource_id": "9b651dfd-7d30-402b-972e-212b2c4bfb05",
    "timestamp": "2014-11-03T13:37:46",
    "meter": "image",
    "volume": 1.0,
    "source": "openstack",
    "recorded_at": "2014-11-03T13:37:46.994458",
    "project_id": "2cc3a7bb859b4bacbeab0aa9ca673033",
    "type": "gauge",
    "id": "98b5f258-635e-11e4-8bdd-0025647390c1",
    "unit": "image",
    "resource_metadata": {},
}

METER_URL = '/v2/meters/instance'
METER_URL_DIRECT = '/v2/meters/instance?direct=True'
SECOND_METER_URL = '/v2/meters/image'
SECOND_METER_URL_DIRECT = '/v2/meters/image?direct=True'
SAMPLE_URL = '/v2/samples'
QUERIES = ('q.field=resource_id&q.field=source&q.op=&q.op='
           '&q.type=&q.type=&q.value=foo&q.value=bar')
LIMIT = 'limit=1'

OLD_SAMPLE_FIXTURES = {
    METER_URL: {
        'GET': (
            {},
            [GET_OLD_SAMPLE]
        ),
        'POST': (
            {},
            [CREATE_SAMPLE],
        ),
    },
    METER_URL_DIRECT: {
        'POST': (
            {},
            [CREATE_SAMPLE],
        )
    },
    SECOND_METER_URL: {
        'POST': (
            {},
            [CREATE_LIST_SAMPLE] * 10,
        ),
    },
    SECOND_METER_URL_DIRECT: {
        'POST': (
            {},
            [CREATE_LIST_SAMPLE] * 10,
        )
    },
    '%s?%s' % (METER_URL, QUERIES): {
        'GET': (
            {},
            [],
        ),
    },
    '%s?%s' % (METER_URL, LIMIT): {
        'GET': (
            {},
            [GET_OLD_SAMPLE]
        ),
    }
}
SAMPLE_FIXTURES = {
    SAMPLE_URL: {
        'GET': (
            (),
            [GET_SAMPLE]
        ),
    },
    '%s?%s' % (SAMPLE_URL, QUERIES): {
        'GET': (
            {},
            [],
        ),
    },
    '%s?%s' % (SAMPLE_URL, LIMIT): {
        'GET': (
            {},
            [GET_SAMPLE],
        ),
    },
    '%s/%s' % (SAMPLE_URL, GET_SAMPLE['id']): {
        'GET': (
            {},
            GET_SAMPLE,
        ),
    },
}


class OldSampleManagerTest(utils.BaseTestCase):

    def setUp(self):
        super(OldSampleManagerTest, self).setUp()
        self.http_client = fake_client.FakeHTTPClient(
            fixtures=OLD_SAMPLE_FIXTURES)
        self.api = client.BaseClient(self.http_client)
        self.mgr = ceilometerclient.v2.samples.OldSampleManager(self.api)

    def test_list_by_meter_name(self):
        samples = list(self.mgr.list(meter_name='instance'))
        expect = [
            'GET', '/v2/meters/instance'
        ]
        self.http_client.assert_called(*expect)
        self.assertEqual(1, len(samples))
        self.assertEqual('resource-id', samples[0].resource_id)

    def test_list_by_meter_name_extended(self):
        samples = list(self.mgr.list(meter_name='instance',
                                     q=[
                                         {"field": "resource_id",
                                          "value": "foo"},
                                         {"field": "source",
                                          "value": "bar"},
                                     ]))
        expect = ['GET', '%s?%s' % (METER_URL, QUERIES)]
        self.http_client.assert_called(*expect)
        self.assertEqual(0, len(samples))

    def test_create(self):
        sample = self.mgr.create(**CREATE_SAMPLE)
        expect = [
            'POST', '/v2/meters/instance'
        ]
        self.http_client.assert_called(*expect, body=[CREATE_SAMPLE])
        self.assertIsNotNone(sample)

    def test_create_directly(self):
        sample = self.mgr.create(direct=True, **CREATE_SAMPLE)
        expect = [
            'POST', '/v2/meters/instance?direct=True'
        ]
        self.http_client.assert_called(*expect, body=[CREATE_SAMPLE])
        self.assertIsNotNone(sample)

    def test_create_list(self):
        test_samples = [CREATE_LIST_SAMPLE] * 10
        samples = self.mgr.create_list(test_samples)
        expect = [
            'POST', '/v2/meters/image'
        ]
        self.http_client.assert_called(*expect, body=test_samples)
        self.assertEqual(10, len(samples))

    def test_create_list_directly(self):
        test_samples = [CREATE_LIST_SAMPLE] * 10
        samples = self.mgr.create_list(test_samples, direct=True)
        expect = [
            'POST', '/v2/meters/image?direct=True'
        ]
        self.http_client.assert_called(*expect, body=test_samples)
        self.assertEqual(10, len(samples))

    def test_limit(self):
        samples = list(self.mgr.list(meter_name='instance', limit=1))
        expect = ['GET', '/v2/meters/instance?limit=1']
        self.http_client.assert_called(*expect)
        self.assertEqual(1, len(samples))


class SampleManagerTest(utils.BaseTestCase):

    def setUp(self):
        super(SampleManagerTest, self).setUp()
        self.http_client = fake_client.FakeHTTPClient(
            fixtures=SAMPLE_FIXTURES)
        self.api = client.BaseClient(self.http_client)
        self.mgr = ceilometerclient.v2.samples.SampleManager(self.api)

    def test_sample_list(self):
        samples = list(self.mgr.list())
        expect = [
            'GET', '/v2/samples'
        ]
        self.http_client.assert_called(*expect)
        self.assertEqual(1, len(samples))
        self.assertEqual('9b651dfd-7d30-402b-972e-212b2c4bfb05',
                         samples[0].resource_id)

    def test_sample_list_with_queries(self):
        queries = [
            {"field": "resource_id",
             "value": "foo"},
            {"field": "source",
             "value": "bar"},
        ]
        samples = list(self.mgr.list(q=queries))
        expect = ['GET', '%s?%s' % (SAMPLE_URL, QUERIES)]
        self.http_client.assert_called(*expect)
        self.assertEqual(0, len(samples))

    def test_sample_list_with_limit(self):
        samples = list(self.mgr.list(limit=1))
        expect = ['GET', '/v2/samples?limit=1']
        self.http_client.assert_called(*expect)
        self.assertEqual(1, len(samples))

    def test_sample_get(self):
        sample = self.mgr.get(GET_SAMPLE['id'])
        expect = ['GET', '/v2/samples/' + GET_SAMPLE['id']]
        self.http_client.assert_called(*expect)
        self.assertEqual(GET_SAMPLE, sample.to_dict())



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from ceilometerclient.tests.unit import utils
from ceilometerclient.v2 import options


class BuildUrlTest(utils.BaseTestCase):

    def test_one(self):
        url = options.build_url('/', [{'field': 'this',
                                       'op': 'gt',
                                       'value': 43}])
        self.assertEqual(url, '/?q.field=this&q.op=gt&q.type=&q.value=43')

    def test_two(self):
        url = options.build_url('/', [{'field': 'this',
                                       'op': 'gt',
                                       'value': 43},
                                      {'field': 'that',
                                       'op': 'lt',
                                       'value': 88}])
        ops = 'q.op=gt&q.op=lt'
        vals = 'q.value=43&q.value=88'
        types = 'q.type=&q.type='
        fields = 'q.field=this&q.field=that'
        self.assertEqual(url, '/?%s&%s&%s&%s' % (fields, ops, types, vals))

    def test_default_op(self):
        url = options.build_url('/', [{'field': 'this',
                                       'value': 43}])
        self.assertEqual(url, '/?q.field=this&q.op=&q.type=&q.value=43')

    def test_one_param(self):
        url = options.build_url('/', None, ['period=60'])
        self.assertEqual(url, '/?period=60')

    def test_two_params(self):
        url = options.build_url('/', None, ['period=60',
                                            'others=value'])
        self.assertEqual(url, '/?period=60&others=value')

    def test_with_data_type(self):
        url = options.build_url('/', [{'field': 'f1',
                                       'value': '10',
                                       'type': 'integer'}])

        self.assertEqual('/?q.field=f1&q.op=&q.type=integer&q.value=10', url)


class CliTest(utils.BaseTestCase):

    def test_one(self):
        ar = options.cli_to_array('this<=34')
        self.assertEqual(ar, [{'field': 'this', 'op': 'le',
                               'value': '34', 'type': ''}])

    def test_two(self):
        ar = options.cli_to_array('this<=34;that!=foo')
        self.assertEqual(ar, [{'field': 'this', 'op': 'le',
                               'value': '34', 'type': ''},
                              {'field': 'that', 'op': 'ne',
                               'value': 'foo', 'type': ''}])

    def test_negative(self):
        ar = options.cli_to_array('this>=-783')
        self.assertEqual(ar, [{'field': 'this', 'op': 'ge',
                               'value': '-783', 'type': ''}])

    def test_float(self):
        ar = options.cli_to_array('this<=283.347')
        self.assertEqual(ar, [{'field': 'this',
                               'op': 'le', 'value': '283.347',
                               'type': ''}])

    def test_comma(self):
        ar = options.cli_to_array('this=2.4,fooo=doof')
        self.assertEqual([{'field': 'this',
                           'op': 'eq',
                           'value': '2.4,fooo=doof',
                           'type': ''}],
                         ar)

    def test_special_character(self):
        ar = options.cli_to_array('key~123=value!123')
        self.assertEqual([{'field': 'key~123',
                           'op': 'eq',
                           'value': 'value!123',
                           'type': ''}],
                         ar)

    def _do_test_typed_float_op(self, op, op_str):
        ar = options.cli_to_array('that%sfloat::283.347' % op)
        self.assertEqual([{'field': 'that',
                           'type': 'float',
                           'value': '283.347',
                           'op': op_str}],
                         ar)

    def test_typed_float_eq(self):
        self._do_test_typed_float_op('<', 'lt')

    def test_typed_float_le(self):
        self._do_test_typed_float_op('<=', 'le')

    def test_typed_string_whitespace(self):
        ar = options.cli_to_array('state=string::insufficient data')
        self.assertEqual([{'field': 'state',
                           'op': 'eq',
                           'type': 'string',
                           'value': 'insufficient data'}],
                         ar)

    def test_typed_string_whitespace_complex(self):
        ar = options.cli_to_array(
            'that>=float::99.9999;state=string::insufficient data'
        )
        self.assertEqual([{'field': 'that',
                           'op': 'ge',
                           'type': 'float',
                           'value': '99.9999'},
                          {'field': 'state',
                           'op': 'eq',
                           'type': 'string',
                           'value': 'insufficient data'}],
                         ar)

    def test_invalid_operator(self):
        self.assertRaises(ValueError, options.cli_to_array,
                          'this=2.4;fooo-doof')

    def test_with_dot(self):
        ar = options.cli_to_array('metadata.this<=34')
        self.assertEqual(ar, [{'field': 'metadata.this',
                               'op': 'le', 'value': '34',
                               'type': ''}])

    def test_single_char_field_or_value(self):
        ar = options.cli_to_array('m<=34;large.thing>s;x!=y')
        self.assertEqual([{'field': 'm',
                           'op': 'le',
                           'value': '34',
                           'type': ''},
                          {'field': 'large.thing',
                           'op': 'gt',
                           'value': 's',
                           'type': ''},
                          {'field': 'x',
                           'op': 'ne',
                           'value': 'y',
                           'type': ''}],
                         ar)

    def test_without_data_type(self):
        ar = options.cli_to_array('hostname=localhost')
        self.assertEqual(ar, [{'field': 'hostname',
                               'op': 'eq',
                               'value': 'localhost',
                               'type': ''}])

    def test_with_string_data_type(self):
        ar = options.cli_to_array('hostname=string::localhost')
        self.assertEqual(ar, [{'field': 'hostname',
                               'op': 'eq',
                               'type': 'string',
                               'value': 'localhost'}])

    def test_with_int_data_type(self):
        ar = options.cli_to_array('port=integer::1234')
        self.assertEqual(ar, [{'field': 'port',
                               'op': 'eq',
                               'type': 'integer',
                               'value': '1234'}])

    def test_with_bool_data_type(self):
        ar = options.cli_to_array('port=boolean::true')
        self.assertEqual(ar, [{'field': 'port',
                               'op': 'eq',
                               'type': 'boolean',
                               'value': 'true'}])

    def test_with_float_data_type(self):
        ar = options.cli_to_array('average=float::1234.5678')
        self.assertEqual(ar, [{'field': 'average',
                               'op': 'eq',
                               'type': 'float',
                               'value': '1234.5678'}])

    def test_with_datetime_data_type(self):
        ar = options.cli_to_array('timestamp=datetime::sometimestamp')
        self.assertEqual(ar, [{'field': 'timestamp',
                               'op': 'eq',
                               'type': 'datetime',
                               'value': 'sometimestamp'}])

    def test_with_incorrect_type(self):
        ar = options.cli_to_array('timestamp=invalid::sometimestamp')
        self.assertEqual(ar, [{'field': 'timestamp',
                               'op': 'eq',
                               'type': '',
                               'value': 'invalid::sometimestamp'}])

    def test_with_single_colon(self):
        ar = options.cli_to_array('timestamp=datetime:sometimestamp')
        self.assertEqual(ar, [{'field': 'timestamp',
                               'op': 'eq',
                               'type': '',
                               'value': 'datetime:sometimestamp'}])

    def test_missing_key(self):
        self.assertRaises(ValueError, options.cli_to_array,
                          'average=float::1234.0;>=string::hello')

    def test_missing_value(self):
        self.assertRaises(ValueError, options.cli_to_array,
                          'average=float::1234.0;house>=')

    def test_timestamp_value(self):
        ar = options.cli_to_array(
            'project=cow;timestamp>=datetime::2014-03-11T16:02:58'
        )
        self.assertEqual([{'field': 'project',
                           'op': 'eq',
                           'type': '',
                           'value': 'cow'},
                          {'field': 'timestamp',
                           'op': 'ge',
                           'type': 'datetime',
                           'value': '2014-03-11T16:02:58'}],
                         ar)

    def test_with_whitespace(self):
        ar = options.cli_to_array('start_timestamp= 2015-01-01T00:00:00;'
                                  ' end_timestamp =2015-06-20T14:01:59 ')

        self.assertEqual([{'field': 'start_timestamp',
                           'op': 'eq',
                           'type': '',
                           'value': '2015-01-01T00:00:00'},
                          {'field': 'end_timestamp',
                           'op': 'eq',
                           'type': '',
                           'value': '2015-06-20T14:01:59'}],
                         ar)



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ceilometerclient.common import base


class EventType(base.Resource):
    def __repr__(self):
        return "<Event %s>" % self._info


def object_class_str(mgr, value, loaded):
    return EventType(mgr, {"event_type": value}, loaded)


class EventTypeManager(base.Manager):

    def list(self):
        return self._list('/v2/event_types/', obj_class=object_class_str)



# -*- coding: utf-8 -*-
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import sys

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT = os.path.abspath(os.path.join(BASE_DIR, "..", "..", ".."))

sys.path.insert(0, ROOT)
sys.path.insert(0, BASE_DIR)


def gen_ref(ver, title, names):
    refdir = os.path.join(BASE_DIR, "ref")
    pkg = "ceilometerclient"
    if ver:
        pkg = "%s.%s" % (pkg, ver)
        refdir = os.path.join(refdir, ver)
    if not os.path.exists(refdir):
        os.makedirs(refdir)
    idxpath = os.path.join(refdir, "index.rst")
    with open(idxpath, "w") as idx:
        idx.write(("%(title)s\n"
                   "%(signs)s\n"
                   "\n"
                   ".. toctree::\n"
                   "   :maxdepth: 1\n"
                   "\n") % {"title": title, "signs": "=" * len(title)})
        for name in names:
            idx.write("   %s\n" % name)
            rstpath = os.path.join(refdir, "%s.rst" % name)
            with open(rstpath, "w") as rst:
                rst.write(("%(title)s\n"
                           "%(signs)s\n"
                           "\n"
                           ".. automodule:: %(pkg)s.%(name)s\n"
                           "   :members:\n"
                           "   :undoc-members:\n"
                           "   :show-inheritance:\n"
                           "   :noindex:\n")
                          % {"title": name.capitalize(),
                             "signs": "=" * len(name),
                             "pkg": pkg, "name": name})



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from ceilometerclient.apiclient import client
from ceilometerclient.apiclient import fake_client
from ceilometerclient.tests.unit import utils
import ceilometerclient.v2.resources


fixtures = {
    '/v2/resources?meter_links=0': {
        'GET': (
            {},
            [
                {
                    'resource_id': 'a',
                    'project_id': 'project_bla',
                    'user_id': 'freddy',
                    'metadata': {'zxc_id': 'bla'},
                },
                {
                    'resource_id': 'b',
                    'project_id': 'dig_the_ditch',
                    'user_id': 'joey',
                    'metadata': {'zxc_id': 'foo'},
                },
            ]
        ),
    },
    '/v2/resources?q.field=resource_id&q.op=&q.type=&q.value=a&meter_links=0':
    {
        'GET': (
            {},
            [
                {
                    'resource_id': 'a',
                    'project_id': 'project_bla',
                    'user_id': 'freddy',
                    'metadata': {'zxc_id': 'bla'},
                },
            ]
        ),
    },
    '/v2/resources?meter_links=1': {
        'GET': (
            {},
            [
                {
                    'resource_id': 'c',
                    'project_id': 'project_blah',
                    'user_id': 'fred',
                    'metadata': {'zxc_id': 'blah'},
                },
                {
                    'resource_id': 'd',
                    'project_id': 'bury_the_ditch',
                    'user_id': 'jack',
                    'metadata': {'zxc_id': 'foobar'},
                },
            ]
        ),
    },
    '/v2/resources/a':
    {
        'GET': (
            {},
            {
                'resource_id': 'a',
                'project_id': 'project_bla',
                'user_id': 'freddy',
                'metadata': {'zxc_id': 'bla'},
            },
        ),
    },
}


class ResourceManagerTest(utils.BaseTestCase):

    def setUp(self):
        super(ResourceManagerTest, self).setUp()
        self.http_client = fake_client.FakeHTTPClient(fixtures=fixtures)
        self.api = client.BaseClient(self.http_client)
        self.mgr = ceilometerclient.v2.resources.ResourceManager(self.api)

    def test_list_all(self):
        resources = list(self.mgr.list())
        expect = [
            'GET', '/v2/resources?meter_links=0'
        ]
        self.http_client.assert_called(*expect)
        self.assertEqual(2, len(resources))
        self.assertEqual('a', resources[0].resource_id)
        self.assertEqual('b', resources[1].resource_id)

    def test_list_all_with_links_enabled(self):
        resources = list(self.mgr.list(links=True))
        expect = [
            'GET', '/v2/resources?meter_links=1'
        ]
        self.http_client.assert_called(*expect)
        self.assertEqual(2, len(resources))
        self.assertEqual('c', resources[0].resource_id)
        self.assertEqual('d', resources[1].resource_id)

    def test_list_one(self):
        resource = self.mgr.get(resource_id='a')
        expect = [
            'GET', '/v2/resources/a'
        ]
        self.http_client.assert_called(*expect)
        self.assertIsNotNone(resource)
        self.assertEqual('a', resource.resource_id)

    def test_list_by_query(self):
        resources = list(self.mgr.list(q=[{"field": "resource_id",
                                           "value": "a"},
                                          ]))
        expect = [
            'GET', '/v2/resources?q.field=resource_id&q.op='
            '&q.type=&q.value=a&meter_links=0'
        ]
        self.http_client.assert_called(*expect)
        self.assertEqual(1, len(resources))
        self.assertEqual('a', resources[0].resource_id)

    def test_get_from_resource_class(self):
        resource = self.mgr.get(resource_id='a')
        self.assertIsNotNone(resource)
        resource.get()
        expect = [
            'GET', '/v2/resources/a'
        ]
        self.http_client.assert_called(*expect, pos=0)
        self.http_client.assert_called(*expect, pos=1)
        self.assertEqual('a', resource.resource_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import time

from keystoneauth1 import adapter
from keystoneauth1 import discover
from keystoneauth1 import exceptions as ka_exc
from keystoneauth1.identity import v2 as v2_auth
from keystoneauth1.identity import v3 as v3_auth
from keystoneauth1 import session
from oslo_utils import importutils
from oslo_utils import strutils
import six.moves.urllib.parse as urlparse

from ceilometerclient.apiclient import auth
from ceilometerclient.apiclient import client
from ceilometerclient.apiclient import exceptions
from ceilometerclient import exc


def _discover_auth_versions(session, auth_url):
    # discover the API versions the server is supporting based on the
    # given URL
    v2_auth_url = None
    v3_auth_url = None
    try:
        ks_discover = discover.Discover(session=session, url=auth_url)
        v2_auth_url = ks_discover.url_for('2.0')
        v3_auth_url = ks_discover.url_for('3.0')
    except ka_exc.DiscoveryFailure:
        raise
    except exceptions.ClientException:
        # Identity service may not support discovery. In that case,
        # try to determine version from auth_url
        url_parts = urlparse.urlparse(auth_url)
        (scheme, netloc, path, params, query, fragment) = url_parts
        path = path.lower()
        if path.startswith('/v3'):
            v3_auth_url = auth_url
        elif path.startswith('/v2'):
            v2_auth_url = auth_url
        else:
            raise exc.CommandError('Unable to determine the Keystone '
                                   'version to authenticate with '
                                   'using the given auth_url.')
    return v2_auth_url, v3_auth_url


def _get_keystone_session(**kwargs):
    # TODO(fabgia): the heavy lifting here should be really done by Keystone.
    # Unfortunately Keystone does not support a richer method to perform
    # discovery and return a single viable URL. A bug against Keystone has
    # been filed: https://bugs.launchpad.net/python-keystoneclient/+bug/1330677

    # first create a Keystone session
    cacert = kwargs.pop('cacert', None)
    cert = kwargs.pop('cert', None)
    key = kwargs.pop('key', None)
    insecure = kwargs.pop('insecure', False)
    auth_url = kwargs.pop('auth_url', None)
    project_id = kwargs.pop('project_id', None)
    project_name = kwargs.pop('project_name', None)
    token = kwargs['token']
    timeout = kwargs.get('timeout')

    if insecure:
        verify = False
    else:
        verify = cacert or True

    if cert and key:
        # passing cert and key together is deprecated in favour of the
        # requests lib form of having the cert and key as a tuple
        cert = (cert, key)

    # create the keystone client session
    ks_session = session.Session(verify=verify, cert=cert, timeout=timeout)
    v2_auth_url, v3_auth_url = _discover_auth_versions(ks_session, auth_url)
    username = kwargs.pop('username', None)
    user_id = kwargs.pop('user_id', None)
    user_domain_name = kwargs.pop('user_domain_name', None)
    user_domain_id = kwargs.pop('user_domain_id', None)
    project_domain_name = kwargs.pop('project_domain_name', None)
    project_domain_id = kwargs.pop('project_domain_id', None)
    if v3_auth_url:
        if not user_domain_name:
            user_domain_name = 'Default'
        if not project_domain_name:
            project_domain_name = 'Default'
    auth = None

    use_domain = (user_domain_id or user_domain_name or
                  project_domain_id or project_domain_name)
    use_v3 = v3_auth_url and (use_domain or (not v2_auth_url))
    use_v2 = v2_auth_url and not use_domain

    if use_v3 and token:
        auth = v3_auth.Token(
            v3_auth_url,
            token=token,
            project_name=project_name,
            project_id=project_id,
            project_domain_name=project_domain_name,
            project_domain_id=project_domain_id)
    elif use_v2 and token:
        auth = v2_auth.Token(
            v2_auth_url,
            token=token,
            tenant_id=project_id,
            tenant_name=project_name)
    elif use_v3:
        # the auth_url as v3 specified
        # e.g. http://no.where:5000/v3
        # Keystone will return only v3 as viable option
        auth = v3_auth.Password(
            v3_auth_url,
            username=username,
            password=kwargs.pop('password', None),
            user_id=user_id,
            user_domain_name=user_domain_name,
            user_domain_id=user_domain_id,
            project_name=project_name,
            project_id=project_id,
            project_domain_name=project_domain_name,
            project_domain_id=project_domain_id)
    elif use_v2:
        # the auth_url as v2 specified
        # e.g. http://no.where:5000/v2.0
        # Keystone will return only v2 as viable option
        auth = v2_auth.Password(
            v2_auth_url,
            username,
            kwargs.pop('password', None),
            tenant_id=project_id,
            tenant_name=project_name)

    else:
        raise exc.CommandError('Unable to determine the Keystone version '
                               'to authenticate with using the given '
                               'auth_url.')

    ks_session.auth = auth
    return ks_session


def _get_endpoint(ks_session, **kwargs):
    """Get an endpoint using the provided keystone session."""

    # set service specific endpoint types
    endpoint_type = kwargs.get('endpoint_type') or 'publicURL'
    service_type = kwargs.get('service_type') or 'metering'

    endpoint = ks_session.get_endpoint(service_type=service_type,
                                       interface=endpoint_type,
                                       region_name=kwargs.get('region_name'))

    return endpoint


class AuthPlugin(auth.BaseAuthPlugin):
    opt_names = ['tenant_id', 'region_name', 'auth_token',
                 'service_type', 'endpoint_type', 'cacert',
                 'auth_url', 'insecure', 'cert_file', 'key_file',
                 'cert', 'key', 'tenant_name', 'project_name',
                 'project_id', 'project_domain_id', 'project_domain_name',
                 'user_id', 'user_domain_id', 'user_domain_name',
                 'password', 'username', 'endpoint']

    def __init__(self, auth_system=None, **kwargs):
        self.opt_names.extend(self.common_opt_names)
        super(AuthPlugin, self).__init__(auth_system, **kwargs)
        # NOTE(sileht): backward compat
        if self.opts.get('auth_token') and not self.opts.get('token'):
            self.opts['token'] = self.opts.get('auth_token')

    def _do_authenticate(self, http_client):
        token = self.opts.get('token')
        endpoint = self.opts.get('endpoint')
        if not (endpoint and token):
            ks_kwargs = self._get_ks_kwargs(http_timeout=http_client.timeout)
            ks_session = _get_keystone_session(**ks_kwargs)
            if not token:
                token = lambda: ks_session.get_token()
            if not endpoint:
                endpoint = _get_endpoint(ks_session, **ks_kwargs)
        self.opts['token'] = token
        self.opts['endpoint'] = endpoint

    def _get_ks_kwargs(self, http_timeout):
        project_id = (self.opts.get('project_id') or
                      self.opts.get('tenant_id'))
        project_name = (self.opts.get('project_name') or
                        self.opts.get('tenant_name'))
        token = self.opts.get('token')
        ks_kwargs = {
            'username': self.opts.get('username'),
            'password': self.opts.get('password'),
            'user_id': self.opts.get('user_id'),
            'user_domain_id': self.opts.get('user_domain_id'),
            'user_domain_name': self.opts.get('user_domain_name'),
            'project_id': project_id,
            'project_name': project_name,
            'project_domain_name': self.opts.get('project_domain_name'),
            'project_domain_id': self.opts.get('project_domain_id'),
            'auth_url': self.opts.get('auth_url'),
            'cacert': self.opts.get('cacert'),
            'cert': self.opts.get('cert'),
            'key': self.opts.get('key'),
            'insecure': strutils.bool_from_string(
                self.opts.get('insecure')),
            'endpoint_type': self.opts.get('endpoint_type'),
            'service_type': self.opts.get('service_type'),
            'region_name': self.opts.get('region_name'),
            'timeout': http_timeout,
            'token': token() if callable(token) else token,
        }
        return ks_kwargs

    def token_and_endpoint(self, endpoint_type, service_type):
        token = self.opts.get('token')
        if callable(token):
            token = token()
        return token, self.opts.get('endpoint')

    def sufficient_options(self):
        """Check if all required options are present.

        :raises: AuthPluginOptionsMissing
        """
        has_token = self.opts.get('token')
        has_project_domain_or_tenant = (self.opts.get('project_id') or
                                        (self.opts.get('project_name') and
                                        (self.opts.get('user_domain_name') or
                                         self.opts.get('user_domain_id'))) or
                                        (self.opts.get('tenant_id') or
                                         self.opts.get('tenant_name')))
        has_credential = (self.opts.get('username')
                          and has_project_domain_or_tenant
                          and self.opts.get('password')
                          and self.opts.get('auth_url'))
        missing = not (has_token or has_credential)
        if missing:
            missing_opts = []
            opts = ['token', 'endpoint', 'username', 'password', 'auth_url',
                    'tenant_id', 'tenant_name']
            for opt in opts:
                if not self.opts.get(opt):
                    missing_opts.append(opt)
            raise exceptions.AuthPluginOptionsMissing(missing_opts)


def _adjust_kwargs(kwargs):
    client_kwargs = {
        'username': kwargs.get('os_username'),
        'password': kwargs.get('os_password'),
        'tenant_id': kwargs.get('os_tenant_id'),
        'tenant_name': kwargs.get('os_tenant_name'),
        'auth_url': kwargs.get('os_auth_url'),
        'region_name': kwargs.get('os_region_name'),
        'service_type': kwargs.get('os_service_type'),
        'endpoint_type': kwargs.get('os_endpoint_type'),
        'insecure': kwargs.get('os_insecure'),
        'cacert': kwargs.get('os_cacert'),
        'cert_file': kwargs.get('os_cert'),
        'key_file': kwargs.get('os_key'),
        'token': kwargs.get('os_token') or kwargs.get('os_auth_token'),
        'user_domain_name': kwargs.get('os_user_domain_name'),
        'user_domain_id': kwargs.get('os_user_domain_id'),
        'project_domain_name': kwargs.get('os_project_domain_name'),
        'project_domain_id': kwargs.get('os_project_domain_id'),
    }

    client_kwargs.update(kwargs)
    client_kwargs['token'] = (client_kwargs.get('token') or
                              kwargs.get('token') or
                              kwargs.get('auth_token'))

    timeout = kwargs.get('timeout')
    if timeout is not None:
        timeout = int(timeout)
        if timeout <= 0:
            timeout = None

    insecure = strutils.bool_from_string(client_kwargs.get('insecure'))
    verify = kwargs.get('verify')
    if verify is None:
        if insecure:
            verify = False
        else:
            verify = client_kwargs.get('cacert') or True

    cert = client_kwargs.get('cert_file')
    key = client_kwargs.get('key_file')
    if cert and key:
        cert = cert, key

    client_kwargs.update({'verify': verify, 'cert': cert, 'timeout': timeout})
    return client_kwargs


def Client(version, *args, **kwargs):
    client_kwargs = _adjust_kwargs(kwargs)

    module = importutils.import_versioned_module('ceilometerclient',
                                                 version, 'client')
    client_class = getattr(module, 'Client')
    return client_class(*args, **client_kwargs)


def get_client(version, **kwargs):
    """Get an authenticated client, based on the credentials in the kwargs.

    :param version: the API version to use ('1' or '2')
    :param kwargs: keyword args containing credentials, either:

            * session: a keystoneauth/keystoneclient session object
            * service_type: The default service_type for URL discovery
            * service_name: The default service_name for URL discovery
            * interface: The default interface for URL discovery
                         (Default: public)
            * region_name: The default region_name for URL discovery
            * endpoint_override: Always use this endpoint URL for requests
                                 for this ceiloclient
            * auth: An auth plugin to use instead of the session one
            * user_agent: The User-Agent string to set
                          (Default is python-ceilometer-client)
            * connect_retries: the maximum number of retries that should be
                               attempted for connection errors
            * logger: A logging object

            or (DEPRECATED):

            * os_auth_token: (DEPRECATED) pre-existing token to re-use,
                             use os_token instead
            * os_token: pre-existing token to re-use
            * ceilometer_url: (DEPRECATED) Ceilometer API endpoint,
                              use os_endpoint instead
            * os_endpoint: Ceilometer API endpoint

            or (DEPRECATED):

            * os_username: name of user
            * os_password: user's password
            * os_user_id: user's id
            * os_user_domain_id: the domain id of the user
            * os_user_domain_name: the domain name of the user
            * os_project_id: the user project id
            * os_tenant_id: V2 alternative to os_project_id
            * os_project_name: the user project name
            * os_tenant_name: V2 alternative to os_project_name
            * os_project_domain_name: domain name for the user project
            * os_project_domain_id: domain id for the user project
            * os_auth_url: endpoint to authenticate against
            * os_cert|os_cacert: path of CA TLS certificate
            * os_key: SSL private key
            * os_insecure: allow insecure SSL (no cert verification)
    """
    endpoint = kwargs.get('os_endpoint') or kwargs.get('ceilometer_url')

    return Client(version, endpoint, **kwargs)


def get_auth_plugin(endpoint, **kwargs):
    auth_plugin = AuthPlugin(
        auth_url=kwargs.get('auth_url'),
        service_type=kwargs.get('service_type'),
        token=kwargs.get('token'),
        endpoint_type=kwargs.get('endpoint_type'),
        insecure=kwargs.get('insecure'),
        region_name=kwargs.get('region_name'),
        cacert=kwargs.get('cacert'),
        tenant_id=kwargs.get('project_id') or kwargs.get('tenant_id'),
        endpoint=endpoint,
        username=kwargs.get('username'),
        password=kwargs.get('password'),
        tenant_name=kwargs.get('project_name') or kwargs.get('tenant_name'),
        user_domain_name=kwargs.get('user_domain_name'),
        user_domain_id=kwargs.get('user_domain_id'),
        project_domain_name=kwargs.get('project_domain_name'),
        project_domain_id=kwargs.get('project_domain_id')
    )
    return auth_plugin


LEGACY_OPTS = ('auth_plugin', 'auth_url', 'token', 'insecure', 'cacert',
               'tenant_id', 'project_id', 'username', 'password',
               'project_name', 'tenant_name',
               'user_domain_name', 'user_domain_id',
               'project_domain_name', 'project_domain_id',
               'key_file', 'cert_file', 'verify', 'timeout', 'cert')


def _construct_http_client(**kwargs):
    kwargs = kwargs.copy()
    if kwargs.get('session') is not None:
        # Drop legacy options
        for opt in LEGACY_OPTS:
            kwargs.pop(opt, None)
        # Drop redirect endpoints from kwargs
        kwargs.pop('aodh_endpoint', None)
        kwargs.pop('panko_endpoint', None)

        return SessionClient(
            session=kwargs.pop('session'),
            service_type=kwargs.pop('service_type', 'metering') or 'metering',
            interface=kwargs.pop('interface', kwargs.pop('endpoint_type',
                                                         'publicURL')),
            region_name=kwargs.pop('region_name', None),
            user_agent=kwargs.pop('user_agent', 'python-ceilometerclient'),
            auth=kwargs.get('auth'),
            timings=kwargs.pop('timings', None),
            **kwargs)
    else:
        return client.BaseClient(client.HTTPClient(
            auth_plugin=kwargs.get('auth_plugin'),
            region_name=kwargs.get('region_name'),
            endpoint_type=kwargs.get('endpoint_type'),
            original_ip=kwargs.get('original_ip'),
            verify=kwargs.get('verify'),
            cert=kwargs.get('cert'),
            timeout=kwargs.get('timeout'),
            timings=kwargs.get('timings'),
            keyring_saver=kwargs.get('keyring_saver'),
            debug=kwargs.get('debug'),
            user_agent=kwargs.get('user_agent'),
            http=kwargs.get('http')
        ))


@contextlib.contextmanager
def record_time(times, enabled, *args):
    """Record the time of a specific action.

    :param times: A list of tuples holds time data.
    :type times: list
    :param enabled: Whether timing is enabled.
    :type enabled: bool
    :param args: Other data to be stored besides time data, these args
                 will be joined to a string.
    """
    if not enabled:
        yield
    else:
        start = time.time()
        yield
        end = time.time()
        times.append((' '.join(args), start, end))


class SessionClient(adapter.LegacyJsonAdapter):
    def __init__(self, *args, **kwargs):
        self.times = []
        self.timings = kwargs.pop('timings', False)
        super(SessionClient, self).__init__(*args, **kwargs)

    def request(self, url, method, **kwargs):
        kwargs.setdefault('headers', kwargs.get('headers', {}))
        # NOTE(sileht): The standard call raises errors from
        # keystoneauth, where we need to raise the ceilometerclient errors.
        raise_exc = kwargs.pop('raise_exc', True)
        with record_time(self.times, self.timings, method, url):
            resp, body = super(SessionClient, self).request(url,
                                                            method,
                                                            raise_exc=False,
                                                            **kwargs)

        if raise_exc and resp.status_code >= 400:
            raise exc.from_response(resp, body)
        return resp



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import re
import sys

import fixtures
from keystoneauth1 import session as ks_session
import mock
import six
from testtools import matchers

from ceilometerclient.apiclient import client as api_client
from ceilometerclient import client
from ceilometerclient import exc
from ceilometerclient import shell as ceilometer_shell
from ceilometerclient.tests.unit import utils

FAKE_V2_ENV = {'OS_USERNAME': 'username',
               'OS_PASSWORD': 'password',
               'OS_TENANT_NAME': 'tenant_name',
               'OS_AUTH_URL': 'http://localhost:5000/v2.0'}

FAKE_V3_ENV = {'OS_USERNAME': 'username',
               'OS_PASSWORD': 'password',
               'OS_USER_DOMAIN_NAME': 'domain_name',
               'OS_PROJECT_ID': '1234567890',
               'OS_AUTH_URL': 'http://localhost:5000/v3'}


class ShellTestBase(utils.BaseTestCase):

    @mock.patch('sys.stdout', new=six.StringIO())
    @mock.patch.object(ks_session, 'Session', mock.MagicMock())
    @mock.patch.object(client.client.HTTPClient,
                       'client_request', mock.MagicMock())
    def shell(self, argstr):
        try:
            _shell = ceilometer_shell.CeilometerShell()
            _shell.main(argstr.split())
        except SystemExit:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.assertEqual(0, exc_value.code)

        return sys.stdout.getvalue()

    # Patch os.environ to avoid required auth info.
    def make_env(self, env_version, exclude=None):
        env = dict((k, v) for k, v in env_version.items() if k != exclude)
        self.useFixture(fixtures.MonkeyPatch('os.environ', env))


class ShellHelpTest(ShellTestBase):
    RE_OPTIONS = re.DOTALL | re.MULTILINE

    def test_help_unknown_command(self):
        self.assertRaises(exc.CommandError, self.shell, 'help foofoo')

    def test_help(self):
        required = [
            '.*?^usage: ceilometer',
            '.*?^See "ceilometer help COMMAND" '
            'for help on a specific command',
        ]
        for argstr in ['--help', 'help']:
            help_text = self.shell(argstr)
            for r in required:
                self.assertThat(help_text,
                                matchers.MatchesRegex(r,
                                                      self.RE_OPTIONS))

    def test_help_on_subcommand(self):
        required = [
            '.*?^usage: ceilometer meter-list',
            ".*?^List the user's meter",
        ]
        argstrings = [
            'help meter-list',
        ]
        for argstr in argstrings:
            help_text = self.shell(argstr)
            for r in required:
                self.assertThat(help_text,
                                matchers.MatchesRegex(r, self.RE_OPTIONS))

    def test_get_base_parser(self):
        standalone_shell = ceilometer_shell.CeilometerShell()
        parser = standalone_shell.get_base_parser()
        self.assertEqual(600, parser.get_default('timeout'))


class ShellBashCompletionTest(ShellTestBase):

    def test_bash_completion(self):
        completion_commands = self.shell("bash-completion")
        options = completion_commands.split(' ')
        self.assertNotIn('bash_completion', options)
        for option in options:
            self.assertThat(option,
                            matchers.MatchesRegex(r'[a-z0-9-]'))


class ShellKeystoneV2Test(ShellTestBase):

    @mock.patch.object(ks_session, 'Session')
    @mock.patch('ceilometerclient.v2.client.Client._get_redirect_client',
                mock.Mock(return_value=None))
    def test_debug_switch_raises_error(self, mock_ksclient):
        mock_ksclient.side_effect = exc.HTTPUnauthorized
        self.make_env(FAKE_V2_ENV)
        args = ['--debug', 'event-list']
        self.assertRaises(exc.CommandError, ceilometer_shell.main, args)

    @mock.patch.object(ks_session, 'Session')
    @mock.patch('ceilometerclient.v2.client.Client._get_redirect_client',
                mock.Mock(return_value=None))
    def test_dash_d_switch_raises_error(self, mock_ksclient):
        mock_ksclient.side_effect = exc.CommandError("FAIL")
        self.make_env(FAKE_V2_ENV)
        args = ['-d', 'event-list']
        self.assertRaises(exc.CommandError, ceilometer_shell.main, args)

    @mock.patch('sys.stderr')
    @mock.patch.object(ks_session, 'Session')
    def test_no_debug_switch_no_raises_errors(self, mock_ksclient, __):
        mock_ksclient.side_effect = exc.HTTPUnauthorized("FAIL")
        self.make_env(FAKE_V2_ENV)
        args = ['event-list']
        self.assertRaises(SystemExit, ceilometer_shell.main, args)


class ShellKeystoneV3Test(ShellTestBase):

    @mock.patch.object(ks_session, 'Session')
    @mock.patch('ceilometerclient.v2.client.Client._get_redirect_client',
                mock.Mock(return_value=None))
    def test_debug_switch_raises_error(self, mock_ksclient):
        mock_ksclient.side_effect = exc.HTTPUnauthorized
        self.make_env(FAKE_V3_ENV)
        args = ['--debug', 'event-list']
        self.assertRaises(exc.CommandError, ceilometer_shell.main, args)

    @mock.patch.object(ks_session, 'Session')
    def test_dash_d_switch_raises_error(self, mock_ksclient):
        mock_ksclient.side_effect = exc.CommandError("FAIL")
        self.make_env(FAKE_V3_ENV)
        args = ['-d', 'event-list']
        self.assertRaises(exc.CommandError, ceilometer_shell.main, args)

    @mock.patch('sys.stderr')
    @mock.patch.object(ks_session, 'Session')
    def test_no_debug_switch_no_raises_errors(self, mock_ksclient, __):
        mock_ksclient.side_effect = exc.HTTPUnauthorized("FAIL")
        self.make_env(FAKE_V3_ENV)
        args = ['event-list']
        self.assertRaises(SystemExit, ceilometer_shell.main, args)


class ShellTimeoutTest(ShellTestBase):

    @mock.patch('sys.stderr', new=six.StringIO())
    def _test_timeout(self, timeout, expected_msg):
        args = ['--timeout', timeout, 'alarm-list']
        self.assertRaises(SystemExit, ceilometer_shell.main, args)
        self.assertEqual(expected_msg, sys.stderr.getvalue().splitlines()[-1])

    def test_timeout_invalid_value(self):
        expected_msg = ('ceilometer: error: argument --timeout: '
                        'abc must be an integer')
        self._test_timeout('abc', expected_msg)

    def test_timeout_negative_value(self):
        expected_msg = ('ceilometer: error: argument --timeout: '
                        '-1 must be greater than 0')
        self._test_timeout('-1', expected_msg)

    def test_timeout_float_value(self):
        expected_msg = ('ceilometer: error: argument --timeout: '
                        '1.5 must be an integer')
        self._test_timeout('1.5', expected_msg)

    def test_timeout_zero(self):
        expected_msg = ('ceilometer: error: argument --timeout: '
                        '0 must be greater than 0')
        self._test_timeout('0', expected_msg)

    @mock.patch.object(ks_session, 'Session')
    @mock.patch('ceilometerclient.v2.client.Client._get_redirect_client',
                mock.Mock(return_value=None))
    def test_timeout_keystone_session(self, mocked_session):
        mocked_session.side_effect = exc.HTTPUnauthorized("FAIL")
        self.make_env(FAKE_V2_ENV)
        args = ['--debug', '--timeout', '5', 'alarm-list']
        self.assertRaises(exc.CommandError, ceilometer_shell.main, args)
        args, kwargs = mocked_session.call_args
        self.assertEqual(5, kwargs.get('timeout'))


class ShellInsecureTest(ShellTestBase):

    @mock.patch.object(api_client, 'HTTPClient')
    @mock.patch('ceilometerclient.v2.client.Client._get_redirect_client',
                mock.Mock(return_value=None))
    def test_insecure_true_ceilometer(self, mocked_client):
        self.make_env(FAKE_V2_ENV)
        args = ['--debug', '--os-insecure', 'true', 'alarm-list']
        self.assertIsNone(ceilometer_shell.main(args))
        args, kwargs = mocked_client.call_args
        self.assertFalse(kwargs.get('verify'))

    @mock.patch.object(ks_session, 'Session')
    @mock.patch('ceilometerclient.v2.client.Client._get_redirect_client',
                mock.Mock(return_value=None))
    def test_insecure_true_keystone(self, mocked_session):
        mocked_session.side_effect = exc.HTTPUnauthorized("FAIL")
        self.make_env(FAKE_V2_ENV)
        args = ['--debug', '--os-insecure', 'true', 'alarm-list']
        self.assertRaises(exc.CommandError, ceilometer_shell.main, args)
        args, kwargs = mocked_session.call_args
        self.assertFalse(kwargs.get('verify'))

    @mock.patch.object(api_client, 'HTTPClient')
    @mock.patch('ceilometerclient.v2.client.Client._get_redirect_client',
                mock.Mock(return_value=None))
    def test_insecure_false_ceilometer(self, mocked_client):
        self.make_env(FAKE_V2_ENV)
        args = ['--debug', '--os-insecure', 'false', 'alarm-list']
        self.assertIsNone(ceilometer_shell.main(args))
        args, kwargs = mocked_client.call_args
        self.assertTrue(kwargs.get('verify'))

    @mock.patch.object(ks_session, 'Session')
    @mock.patch('ceilometerclient.v2.client.Client._get_redirect_client',
                mock.Mock(return_value=None))
    def test_insecure_false_keystone(self, mocked_session):
        mocked_session.side_effect = exc.HTTPUnauthorized("FAIL")
        self.make_env(FAKE_V2_ENV)
        args = ['--debug', '--os-insecure', 'false', 'alarm-list']
        self.assertRaises(exc.CommandError, ceilometer_shell.main, args)
        args, kwargs = mocked_session.call_args
        self.assertTrue(kwargs.get('verify'))


class ShellEndpointTest(ShellTestBase):

    @mock.patch('ceilometerclient.v2.client.Client')
    def _test_endpoint_and_token(self, token_name, endpoint_name, mocked):
        args = ['--debug', token_name, 'fake-token',
                endpoint_name, 'http://fake-url', 'alarm-list']
        self.assertIsNone(ceilometer_shell.main(args))
        args, kwargs = mocked.call_args
        self.assertEqual('http://fake-url', kwargs.get('endpoint'))
        self.assertEqual('fake-token', kwargs.get('token'))

    def test_endpoint_and_token(self):
        self._test_endpoint_and_token('--os-auth-token', '--ceilometer-url')
        self._test_endpoint_and_token('--os-auth-token', '--os-endpoint')
        self._test_endpoint_and_token('--os-token', '--ceilometer-url')
        self._test_endpoint_and_token('--os-token', '--os-endpoint')


class ShellAlarmUpdateRepeatAction(ShellTestBase):
    @mock.patch('ceilometerclient.v2.alarms.AlarmManager.update')
    @mock.patch('ceilometerclient.v2.client.Client._get_redirect_client',
                mock.Mock())
    def test_repeat_action_not_specified(self, mocked):
        self.make_env(FAKE_V2_ENV)

        def _test(method):
            args = ['--debug', method, '--state', 'alarm', '123']
            ceilometer_shell.main(args)
            args, kwargs = mocked.call_args
            self.assertIsNone(kwargs.get('repeat_actions'))

        _test('alarm-update')
        _test('alarm-threshold-update')
        _test('alarm-combination-update')
        _test('alarm-event-update')



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from ceilometerclient.apiclient import client
from ceilometerclient.apiclient import fake_client
from ceilometerclient.tests.unit import utils
import ceilometerclient.v2.event_types


fixtures = {
    '/v2/event_types/': {
        'GET': (
            {},
            ['Foo', 'Bar', 'Sna', 'Fu']
        ),
    }
}


class EventTypesManagerTest(utils.BaseTestCase):

    def setUp(self):
        super(EventTypesManagerTest, self).setUp()
        self.http_client = fake_client.FakeHTTPClient(fixtures=fixtures)
        self.api = client.BaseClient(self.http_client)
        self.mgr = ceilometerclient.v2.event_types.EventTypeManager(self.api)

    def test_list(self):
        event_types = list(self.mgr.list())
        expect = [
            'GET', '/v2/event_types/'
        ]
        self.http_client.assert_called(*expect)
        self.assertEqual(4, len(event_types))
        self.assertEqual("Foo", event_types[0].event_type)
        self.assertEqual("Bar", event_types[1].event_type)
        self.assertEqual("Sna", event_types[2].event_type)
        self.assertEqual("Fu", event_types[3].event_type)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
execfile(os.path.join("..", "ext", "gen_ref.py"))

project = 'python-ceilometerclient'

gen_ref("", "Client Reference", ["client", "exc"])
gen_ref("v2", "Version 2 API Reference",
        ["meters", "samples", "statistics", "resources", "query", "alarms",
         "events", "event_types", "traits", "trait_descriptions"])

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc', 'oslosphinx' ]

# autodoc generation is a bit aggressive and a nuisance when doing heavy
# text edit cycles.
# execute "export SPHINX_DEBUG=1" in your terminal to disable

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
copyright = u'OpenStack Foundation'

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
#html_theme = 'nature'

# Output file base name for HTML help builder.
htmlhelp_basename = '%sdoc' % project


# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    (
        'index',
        '%s.tex' % project,
        u'%s Documentation' % project,
        u'OpenStack Foundation',
        'manual'
    ),
]



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from six.moves import urllib

OP_LOOKUP = {'!=': 'ne',
             '>=': 'ge',
             '<=': 'le',
             '>': 'gt',
             '<': 'lt',
             '=': 'eq'}

OP_LOOKUP_KEYS = '|'.join(sorted(OP_LOOKUP.keys(), key=len, reverse=True))
OP_SPLIT_RE = re.compile(r'(%s)' % OP_LOOKUP_KEYS)

DATA_TYPE_RE = re.compile(r'^(string|integer|float|datetime|boolean)(::)(.+)$')


def build_url(path, q, params=None):
    """Convert list of dicts and a list of params to query url format.

    This will convert the following:
        "[{field=this,op=le,value=34},
          {field=that,op=eq,value=foo,type=string}],
         ['foo=bar','sna=fu']"
    to:
        "?q.field=this&q.field=that&
          q.op=le&q.op=eq&
          q.type=&q.type=string&
          q.value=34&q.value=foo&
          foo=bar&sna=fu"
    """
    if q:
        query_params = {'q.field': [],
                        'q.value': [],
                        'q.op': [],
                        'q.type': []}

        for query in q:
            for name in ['field', 'op', 'value', 'type']:
                query_params['q.%s' % name].append(query.get(name, ''))

        # Transform the dict to a sequence of two-element tuples in fixed
        # order, then the encoded string will be consistent in Python 2&3.
        new_qparams = sorted(query_params.items(), key=lambda x: x[0])
        path += "?" + urllib.parse.urlencode(new_qparams, doseq=True)

        if params:
            for p in params:
                path += '&%s' % p
    elif params:
        path += '?%s' % params[0]
        for p in params[1:]:
            path += '&%s' % p
    return path


def cli_to_array(cli_query):
    """Convert CLI list of queries to the Python API format.

    This will convert the following:
        "this<=34;that=string::foo"
    to
        "[{field=this,op=le,value=34,type=''},
          {field=that,op=eq,value=foo,type=string}]"

    """

    if cli_query is None:
        return None

    def split_by_op(query):
        """Split a single query string to field, operator, value."""

        def _value_error(message):
            raise ValueError('invalid query %(query)s: missing %(message)s' %
                             {'query': query, 'message': message})

        try:
            field, operator, value = OP_SPLIT_RE.split(query, maxsplit=1)
        except ValueError:
            _value_error('operator')

        if not len(field):
            _value_error('field')

        if not len(value):
            _value_error('value')

        return field.strip(), operator, value.strip()

    def split_by_data_type(query_value):
        frags = DATA_TYPE_RE.match(query_value)

        # The second match is the separator. Return a list without it if
        # a type identifier was found.
        return frags.group(1, 3) if frags else None

    opts = []
    queries = cli_query.split(';')
    for q in queries:
        query = split_by_op(q)
        opt = {}
        opt['field'] = query[0]
        opt['op'] = OP_LOOKUP[query[1]]

        # Allow the data type of the value to be specified via <type>::<value>,
        # where type can be one of integer, string, float, datetime, boolean
        value_frags = split_by_data_type(query[2])
        if not value_frags:
            opt['value'] = query[2]
            opt['type'] = ''
        else:
            opt['type'] = value_frags[0]
            opt['value'] = value_frags[1]
        opts.append(opt)
    return opts



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from ceilometerclient.apiclient import client
from ceilometerclient.apiclient import fake_client
from ceilometerclient.tests.unit import utils
import ceilometerclient.v2.traits


fixtures = {
    '/v2/event_types/Foo/traits/trait_1': {
        'GET': (
            {},
            [
                {'name': 'trait_1',
                 'type': 'datetime',
                 'value': '2014-01-07T17:22:10.925553'},
                {'name': 'trait_1',
                 'type': 'datetime',
                 'value': '2014-01-07T17:23:10.925553'}
            ]
        ),
    }
}


class TraitManagerTest(utils.BaseTestCase):

    def setUp(self):
        super(TraitManagerTest, self).setUp()
        self.http_client = fake_client.FakeHTTPClient(fixtures=fixtures)
        self.api = client.BaseClient(self.http_client)
        self.mgr = ceilometerclient.v2.traits.TraitManager(self.api)

    def test_list(self):
        traits = list(self.mgr.list('Foo', 'trait_1'))
        expect = [
            'GET', '/v2/event_types/Foo/traits/trait_1'
        ]
        self.http_client.assert_called(*expect)
        self.assertEqual(2, len(traits))
        for i, vals in enumerate([('trait_1',
                                   'datetime',
                                   '2014-01-07T17:22:10.925553'),
                                  ('trait_1',
                                   'datetime',
                                   '2014-01-07T17:23:10.925553')]):

            name, type, value = vals
            self.assertEqual(traits[i].name, name)
            self.assertEqual(traits[i].type, type)
            self.assertEqual(traits[i].value, value)
